// Author: Jean Honorio
// E-mail: jhonorio@csail.mit.edu

#include "mex.h"
#include <math.h>
#include <stdlib.h>
#include <time.h>

void mexFunction(int nargout, mxArray *argout[], int nargin, const mxArray *argin[])
{
    static int loaded;
    int derivative, N, t, T, *x, r, i, j, k, l;
    double *W, *b, *P1, *W_ptr, A, exp_fx, Z, *DW, *Db, *DW_ptr;

    if (!loaded) {
        mexPrintf("Setting seed inside MEX\n");
        srand(time(NULL));
        loaded = 1;
    }
    if (nargin != 4)
        mexErrMsgTxt("Four input arguments expected");
    if (nargout != 2)
        mexErrMsgTxt("Two output argument expected");
    N = mxGetM(argin [0]);
    if (mxGetN(argin [0]) != N)
        mexErrMsgTxt("Matrix W is not square");
    W = mxGetPr(argin [0]);
    if (mxGetM(argin [1]) * mxGetN(argin [1]) != N)
        mexErrMsgTxt("Vector b is not the same dimension of W");
    b = mxGetPr(argin [1]);
    if ((x = (int *) calloc(N, sizeof(int))) == NULL)
        mexErrMsgTxt("Insufficient memory");
    T = mxGetScalar(argin [2]);
    if (mxGetM(argin [3]) * mxGetN(argin [3]) != N)
        mexErrMsgTxt("Vector P1 is not the same dimension of W");
    P1 = mxGetPr(argin [3]);
    argout [0] = mxCreateDoubleMatrix(N, N, mxREAL);
    DW = mxGetPr(argout[0]);
    argout [1] = mxCreateDoubleMatrix(N, 1, mxREAL);
    Db = mxGetPr(argout[1]);
    W_ptr = W;
    A = 0;
    for (i = 0; i < N; i++) {
        double s = 0;
        for (j = 0; j < N; j++)
            s += fabs(*(W_ptr++));
        A += s / 2 + fabs(b [i]);
    }
    Z = 0;
    for (t = 0; t < T; t++) {
        double a = 0;
        double *W_ptr = W;
        double gx = 1;
        for (i = 0; i < N; i++)
            if (rand() < P1 [i] * RAND_MAX) {
                x [i] = 1;
                gx *= P1 [i];
            } else {
                x [i] = -1;
                gx *= 1 - P1 [i];
            }
        for (i = 0; i < N; i++) {
            double s = 0;
            int *x_ptr = x;
            for (j = 0; j < N; j++)
                s += *(W_ptr++) * *(x_ptr++);
            a += x [i] * (s / 2 + b [i]);
        }
        exp_fx = exp(a - A) / gx;
        Z += exp_fx;
        DW_ptr = DW;
        for (k = 0; k < N; k++) {
            int *x_ptr = x;
            for (l = 0; l < N; l++)
                *(DW_ptr++) += exp_fx * (x [k] * *(x_ptr++) - (k == l)) / 2;
            Db [k] += exp_fx * x [k];
        }
    }
    free(x);
    DW_ptr = DW;
    for (k = 0; k < N; k++) {
        for (l = 0; l < N; l++)
            *(DW_ptr++) /= Z;
        Db [k] /= Z;
    }
}
